/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.kafka.connect.mirror;

import net.sourceforge.argparse4j.ArgumentParsers;
import net.sourceforge.argparse4j.impl.Arguments;
import net.sourceforge.argparse4j.inf.ArgumentParser;
import net.sourceforge.argparse4j.inf.ArgumentParserException;
import net.sourceforge.argparse4j.inf.Namespace;
import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import org.apache.kafka.common.utils.Exit;
import org.apache.kafka.common.utils.Time;
import org.apache.kafka.common.utils.Utils;
import org.apache.kafka.connect.connector.policy.AllConnectorClientConfigOverridePolicy;
import org.apache.kafka.connect.connector.policy.ConnectorClientConfigOverridePolicy;
import org.apache.kafka.connect.mirror.utils.Retrier;
import org.apache.kafka.connect.mirror.utils.RetryStatus;
import org.apache.kafka.connect.mirror.utils.RingBuffer;
import org.apache.kafka.connect.rest.ConnectRestExtension;
import org.apache.kafka.connect.runtime.Herder;
import org.apache.kafka.connect.runtime.Worker;
import org.apache.kafka.connect.runtime.WorkerConfig;
import org.apache.kafka.connect.runtime.WorkerConfigTransformer;
import org.apache.kafka.connect.runtime.distributed.DistributedConfig;
import org.apache.kafka.connect.runtime.distributed.DistributedHerder;
import org.apache.kafka.connect.runtime.distributed.NotLeaderException;
import org.apache.kafka.connect.runtime.isolation.Plugins;
import org.apache.kafka.connect.runtime.rest.RestServer;
import org.apache.kafka.connect.storage.ConfigBackingStore;
import org.apache.kafka.connect.storage.Converter;
import org.apache.kafka.connect.storage.KafkaConfigBackingStore;
import org.apache.kafka.connect.storage.KafkaOffsetBackingStore;
import org.apache.kafka.connect.storage.KafkaStatusBackingStore;
import org.apache.kafka.connect.storage.StatusBackingStore;
import org.apache.kafka.connect.util.ConnectUtils;

import org.apache.kafka.connect.util.SharedTopicAdmin;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 *  Entry point for "MirrorMaker 2.0".
 *  <p>
 *  MirrorMaker runs a set of Connectors between multiple clusters, in order to replicate data, configuration,
 *  ACL rules, and consumer group state.
 *  </p>
 *  <p>
 *  Configuration is via a top-level "mm2.properties" file, which supports per-cluster and per-replication
 *  sub-configs. Each source->target replication must be explicitly enabled. For example:
 *  </p>
 *  <pre>
 *    clusters = primary, backup
 *    primary.bootstrap.servers = vip1:9092
 *    backup.bootstrap.servers = vip2:9092
 *    primary->backup.enabled = true
 *    backup->primary.enabled = true
 *  </pre>
 *  <p>
 *  Run as follows:
 *  </p>
 *  <pre>
 *    ./bin/connect-mirror-maker.sh mm2.properties
 *  </pre>
 *  <p>
 *  Additional information and example configurations are provided in ./connect/mirror/README.md
 *  </p>
 */
public class MirrorMaker {
    private static final Logger log = LoggerFactory.getLogger(MirrorMaker.class);

    private static final long SHUTDOWN_TIMEOUT_SECONDS = 60L;
    private static final ConnectorClientConfigOverridePolicy CLIENT_CONFIG_OVERRIDE_POLICY =
            new AllConnectorClientConfigOverridePolicy();

    private static final List<Class<?>> CONNECTOR_CLASSES = Arrays.asList(
        MirrorSourceConnector.class,
        MirrorHeartbeatConnector.class,
        MirrorCheckpointConnector.class);

    private final Map<SourceAndTarget, Herder> herders = new ConcurrentHashMap<>();
    private CountDownLatch startLatch;
    private CountDownLatch stopLatch;
    private final AtomicBoolean shutdown = new AtomicBoolean(false);
    private final ShutdownHook shutdownHook;
    private final Time time;
    private final MirrorMakerConfig config;
    private final Set<String> clusters;
    private final Set<SourceAndTarget> herderPairs;

    private final Map<SourceAndTarget, Map<String, String>> workerConfigMap = new HashMap<>();
    private final Map<SourceAndTarget, DistributedConfig>  distributedConfigMap = new HashMap<>();
    private final MirrorMakerStarter mmStarter;


    /**
     * @param config    MM2 configuration from mm2.properties file
     * @param clusters  target clusters for this node. These must match cluster
     *                  aliases as defined in the config. If null or empty list,
     *                  uses all clusters in the config.
     * @param time      time source
     */
    public MirrorMaker(MirrorMakerConfig config, List<String> clusters, Time time) {
        log.debug("Kafka MirrorMaker instance created");
        this.time = time;
        this.config = config;
        if (clusters != null && !clusters.isEmpty()) {
            this.clusters = new HashSet<>(clusters);
        } else {
            // default to all clusters
            this.clusters = config.clusters();
        }
        log.info("Targeting clusters {}", this.clusters);
        this.herderPairs = config.clusterPairs().stream()
            .filter(x -> this.clusters.contains(x.target()))
            .collect(Collectors.toSet());
        if (herderPairs.isEmpty()) {
            throw new IllegalArgumentException("No source->target replication flows.");
        }

        this.herderPairs.forEach(sourceAndTarget -> {
            Map<String, String> workerProps = config.workerConfig(sourceAndTarget);
            Plugins plugins = new Plugins(workerProps);
            plugins.compareAndSwapWithDelegatingLoader();

            workerConfigMap.put(sourceAndTarget, workerProps);
            distributedConfigMap.put(sourceAndTarget, new DistributedConfig(workerProps));
        });

        mmStarter = new MirrorMakerStarter();
        shutdownHook = new ShutdownHook();
    }


    /**
     * @param config    MM2 configuration from mm2.properties file
     * @param clusters  target clusters for this node. These must match cluster
     *                  aliases as defined in the config. If null or empty list,
     *                  uses all clusters in the config.
     * @param time      time source
     */
    public MirrorMaker(Map<String, String> config, List<String> clusters, Time time) {
        this(new MirrorMakerConfig(config), clusters, time);
    }

    public MirrorMaker(Map<String, String> props, List<String> clusters) {
        this(props, clusters, Time.SYSTEM);
    }

    public MirrorMaker(Map<String, String> props) {
        this(props, null);
    }

    public void start() {
        log.info("Kafka MirrorMaker starting with {} herders.", herderPairs.size());
        if (startLatch != null) {
            throw new IllegalStateException("MirrorMaker instance already started");
        }
        startLatch = new CountDownLatch(herderPairs.size());
        stopLatch = new CountDownLatch(herderPairs.size());
        Exit.addShutdownHook("mirror-maker-shutdown-hook", shutdownHook);
        mmStarter.start();
        herderPairs.forEach(mmStarter::start);
        log.info("Kafka MirrorMaker started");
    }

    public void stop() {
        boolean wasShuttingDown = shutdown.getAndSet(true);
        if (!wasShuttingDown) {
            log.info("Kafka MirrorMaker stopping");
            herderPairs.forEach(mmStarter::stop);
            mmStarter.stop();
            log.info("Kafka MirrorMaker stopped.");
        }
    }

    public void awaitStop() {
        try {
            stopLatch.await();
        } catch (InterruptedException e) {
            log.error("Interrupted waiting for MirrorMaker to shutdown");
        }
    }

    private void configureConnector(SourceAndTarget sourceAndTarget, Class<?> connectorClass) {
        checkHerder(sourceAndTarget);
        Map<String, String> connectorProps = config.connectorBaseConfig(sourceAndTarget, connectorClass);
        herders.get(sourceAndTarget)
                .putConnectorConfig(connectorClass.getSimpleName(), connectorProps, true, (e, x) -> {
                    if (e instanceof NotLeaderException) {
                        // No way to determine if the connector is a leader or not beforehand.
                        log.info("Connector {} is a follower. Using existing configuration.", sourceAndTarget);
                    } else {
                        log.info("Connector {} configured.", sourceAndTarget, e);
                    }
                });
    }

    private void checkHerder(SourceAndTarget sourceAndTarget) {
        if (!herders.containsKey(sourceAndTarget)) {
            throw new IllegalArgumentException("No herder for " + sourceAndTarget.toString());
        }
    }

    private void configureConnectors(SourceAndTarget sourceAndTarget) {
        CONNECTOR_CLASSES.forEach(x -> configureConnector(sourceAndTarget, x));
    }

    Herder createHerder(SourceAndTarget sourceAndTarget, URI advertisedUrl) {
        log.info("creating herder for " + sourceAndTarget.toString());
        Plugins plugins = new Plugins(workerConfigMap.get(sourceAndTarget));
        plugins.compareAndSwapWithDelegatingLoader();
        DistributedConfig distributedConfig = distributedConfigMap.get(sourceAndTarget);

        // In Connect, this is a proper URL - but since REST port might be 0, ID would change each restart
        String workerId = advertisedUrl.getHost() + "/" + sourceAndTarget.source() + "__" + sourceAndTarget.target();
        String kafkaClusterId = ConnectUtils.lookupKafkaClusterId(distributedConfig);
        // Create the admin client to be shared by all backing stores for this herder
        Map<String, Object> adminProps = new HashMap<>(distributedConfig.originals());
        ConnectUtils.addMetricsContextProperties(adminProps, distributedConfig, kafkaClusterId);
        SharedTopicAdmin sharedAdmin = new SharedTopicAdmin(adminProps);
        KafkaOffsetBackingStore offsetBackingStore = new KafkaOffsetBackingStore(sharedAdmin);
        offsetBackingStore.configure(distributedConfig);
        Worker worker = new Worker(workerId, time, plugins, distributedConfig, offsetBackingStore, CLIENT_CONFIG_OVERRIDE_POLICY);
        WorkerConfigTransformer configTransformer = worker.configTransformer();
        Converter internalValueConverter = worker.getInternalValueConverter();
        StatusBackingStore statusBackingStore = new KafkaStatusBackingStore(time, internalValueConverter, sharedAdmin);
        statusBackingStore.configure(distributedConfig);
        ConfigBackingStore configBackingStore = new KafkaConfigBackingStore(
                internalValueConverter,
                distributedConfig,
                configTransformer,
                sharedAdmin);
        // Pass the shared admin to the distributed herder as an additional AutoCloseable object that should be closed when the
        // herder is stopped. MirrorMaker has multiple herders, and having the herder own the close responsibility is much easier than
        // tracking the various shared admin objects in this class.
        Herder herder = new DistributedHerder(distributedConfig, time, worker,
                kafkaClusterId, statusBackingStore, configBackingStore,
                advertisedUrl.toString(), CLIENT_CONFIG_OVERRIDE_POLICY, sharedAdmin);

        return herder;
    }

    private class ShutdownHook extends Thread {
        @Override
        public void run() {
            try {
                if (!startLatch.await(SHUTDOWN_TIMEOUT_SECONDS, TimeUnit.SECONDS)) {
                    log.error("Timed out in shutdown hook waiting for MirrorMaker startup to finish. Unable to shutdown cleanly.");
                }
            } catch (InterruptedException e) {
                log.error("Interrupted in shutdown hook while waiting for MirrorMaker startup to finish. Unable to shutdown cleanly.");
            } finally {
                MirrorMaker.this.stop();
            }
        }
    }

    public static void main(String[] args) {
        ArgumentParser parser = ArgumentParsers.newArgumentParser("connect-mirror-maker");
        parser.description("MirrorMaker 2.0 driver");
        parser.addArgument("config").type(Arguments.fileType().verifyCanRead())
            .metavar("mm2.properties").required(true)
            .help("MM2 configuration file.");
        parser.addArgument("--clusters").nargs("+").metavar("CLUSTER").required(false)
            .help("Target cluster to use for this node.");
        Namespace ns;
        try {
            ns = parser.parseArgs(args);
        } catch (ArgumentParserException e) {
            parser.handleError(e);
            Exit.exit(-1);
            return;
        }
        File configFile = ns.get("config");
        List<String> clusters = ns.getList("clusters");
        try {
            log.info("Kafka MirrorMaker initializing ...");

            Properties props = Utils.loadProps(configFile.getPath());
            Map<String, String> config = Utils.propsToStringMap(props);
            MirrorMaker mirrorMaker = new MirrorMaker(config, clusters, Time.SYSTEM);
            
            try {
                mirrorMaker.start();
            } catch (Exception e) {
                log.error("Failed to start MirrorMaker", e);
                mirrorMaker.stop();
                Exit.exit(3);
            }

            mirrorMaker.awaitStop();

        } catch (Throwable t) {
            log.error("Stopping due to error", t);
            Exit.exit(2);
        }
    }

    private class MirrorMakerStarter {
        private final static int HISTORY_SIZE = 5;
        private final static String JMX_JSON_METRICS_REST_EXTENSION = "com.cloudera.dim.kafka.metrics.JmxJsonMetricsRestExtension";
        private final int maxTries;
        private final long delayMs;
        private final MirrorMakerMetrics metrics;
        private ConnectRestExtension jmxJsonRestExtension;
        private final Map<SourceAndTarget, RestServer> restServers = new ConcurrentHashMap<>();
        private final Map<SourceAndTarget, RingBuffer<RetryStatus>> herderStatus = new ConcurrentHashMap<>();
        private final ExecutorService starters = Executors.newFixedThreadPool(
                herderPairs.size(),
                new BasicThreadFactory.Builder()
                    .namingPattern("mmstart-%d")
                    .build());


        private MirrorMakerStarter() {
            this.metrics = new MirrorMakerMetrics(config);
            this.maxTries = config.getInt(MirrorMakerConfig.HERDER_RESTART_NUM_CONFIG);
            this.delayMs = config.getLong(MirrorMakerConfig.HERDER_RESTART_DELAY_CONFIG);
        }

        void start() {
            initMetrics();
            if (config.getBoolean(MirrorMakerConfig.MM_METRICS_SERVLET_ENABLE) &&
                config.originalsStrings().containsKey(WorkerConfig.PLUGIN_PATH_CONFIG)) {
                initMetricsServlet();
            } else {
                log.warn("Not starting metrics servlet, because either metrics servlet is not enabled or plugin.path is not configured.");
            }
        }

        void start(SourceAndTarget sourceAndTarget) {

            // fire and forget
            starters.execute(
                () -> {
                    Optional<?> result = Retrier.of(
                        () -> {
                            // create
                            RestServer rest = computeIfAbsentForMap(restServers, sourceAndTarget, this::createRestServer);
                            Herder herder = computeIfAbsentForMap(herders, sourceAndTarget, st -> createHerder(st, rest.advertisedUrl()));

                            // start
                            try {
                                herder.start();
                            } catch (Exception e) {
                                log.warn("Herder for {} could not start, re-creating it before re-attempting.", sourceAndTarget);
                                herders.remove(sourceAndTarget).stop();
                                throw e;
                            }

                            try {
                                rest.initializeResources(herder);
                            } catch (Exception e) {
                                log.warn("Rest server for {} could not start, re-creating in next run. Also re-creating dependent herder.", sourceAndTarget);
                                herders.remove(sourceAndTarget).stop();
                                restServers.remove(sourceAndTarget).stop();
                                throw e;
                            }

                            //herder URL metrics
                            metrics.herderUrl(sourceAndTarget.toString(), rest.advertisedUrl().toString());

                            // put config
                            log.info("Configuring connectors: {}", sourceAndTarget);
                            configureConnectors(sourceAndTarget);

                            // started
                            startLatch.countDown();

                            return sourceAndTarget;
                        },
                        herderStatus.get(sourceAndTarget),
                        "starting of herder " + sourceAndTarget.toString(),
                        delayMs
                    ).retry(maxTries);

                    if (!result.isPresent()) {
                        log.error("All retry attempts failed to start herder for {}", sourceAndTarget);
                        Exit.exit(1);
                    }
                }
            );
        }

        void stop() {
            /**
             * @see <a href="https://docs.oracle.com/javase/8/docs/api/java/util/concurrent/ExecutorService.html">ExecutorService</a>
             */
            // stopping possibly running retriers
            starters.shutdown();
            try {
                if (!starters.awaitTermination(60, TimeUnit.SECONDS)) {
                    starters.shutdownNow();
                }
            } catch (InterruptedException e) {
                starters.shutdownNow();
                Thread.currentThread().interrupt();
            }

            // stopping metrics
            stopMetricsServlet();
            metrics.close();
        }

        /**
         * Stops running Herders and REST servers identified by their SourceAndTarget key.
         * @param sourceAndTarget
         */
        void stop(SourceAndTarget sourceAndTarget) {
            // metrics
            metrics.removeHerderMetrics(sourceAndTarget.toString());

            // herder
            Optional.ofNullable(herders.get(sourceAndTarget)).ifPresent(Herder::stop);

            // rest
            Optional.ofNullable(restServers.get(sourceAndTarget)).ifPresent(RestServer::stop);

            // stopped
            stopLatch.countDown();
        }

        private void initMetricsServlet() {
            Plugins plugins = new Plugins(config.originalsStrings());
            plugins.compareAndSwapWithDelegatingLoader();
            jmxJsonRestExtension = plugins.newPlugin(JMX_JSON_METRICS_REST_EXTENSION, config, ConnectRestExtension.class);

            // This specific Rest Extension does not use the passed context
            jmxJsonRestExtension.register(null);
        }

        private void stopMetricsServlet() {
            try {
                if (jmxJsonRestExtension != null) {
                    jmxJsonRestExtension.close();
                }
            } catch (IOException e) {
                // should never really be thrown
                log.error("Unexpected IOException during stopping metrics servlet.", e);
            }
        }

        private void initMetrics() {
            herderPairs.forEach(sourceAndTarget -> {
                RingBuffer<RetryStatus> statusList = herderStatus.computeIfAbsent(
                    sourceAndTarget,
                    x -> RingBuffer.of(new RetryStatus[HISTORY_SIZE], RetryStatus.UNKNOWN));

                metrics.herderStatus(sourceAndTarget.toString(), () -> statusList.getLast());
                IntStream.range(0, HISTORY_SIZE).forEach(i -> metrics.herderStatus(sourceAndTarget.toString(), i, () -> statusList.get(i)));
            });
        }

        private RestServer createRestServer(SourceAndTarget sourceAndTarget) {
            DistributedConfig distributedConfig = distributedConfigMap.get(sourceAndTarget);
            RestServer rest = new RestServer(distributedConfig);
            rest.initializeServer();
            return rest;
        }

        private <K, V> V computeIfAbsentForMap(Map<K, V> map, K key, Function<? super K, ? extends V> mappingFunction) {
            V v;
            if (map.containsKey(key)) {
                v = map.get(key);
            } else {
                v = mappingFunction.apply(key);
                map.put(key, v);
            }
            return v;
        }
    }

}
